@@ -136,7 +136,13 @@ struct ndr_print {
 #define LIBNDR_FLAG_ALIGN4       (1<<23)
 #define LIBNDR_FLAG_ALIGN8       (1<<24)
 
-#define LIBNDR_ALIGN_FLAGS (LIBNDR_FLAG_ALIGN2|LIBNDR_FLAG_ALIGN4|LIBNDR_FLAG_ALIGN8)
+#define LIBNDR_ALIGN_FLAGS ( 0        | \
+		LIBNDR_FLAG_NOALIGN   | \
+		LIBNDR_FLAG_REMAINING | \
+		LIBNDR_FLAG_ALIGN2    | \
+		LIBNDR_FLAG_ALIGN4    | \
+		LIBNDR_FLAG_ALIGN8    | \
+		0)
 
 #define LIBNDR_PRINT_ARRAY_HEX   (1<<25)
 #define LIBNDR_PRINT_SET_VALUES  (1<<26)
