@@ -37,6 +37,8 @@
  */
 
 #include "includes.h"
+#include "passdb.h"
+#include "lib/winbind_util.h"
 
 /***************************************************************************
   Default implementations of some functions.
@@ -50,7 +52,7 @@ static NTSTATUS _pdb_wbc_sam_getsampw(struct pdb_methods *methods,
 	if (pwd == NULL)
 		return NT_STATUS_NO_SUCH_USER;
 
-	memset(user, 0, sizeof(user));
+	ZERO_STRUCTP(user);
 
         /* Can we really get away with this little of information */
 	user->methods = methods;
@@ -64,26 +66,26 @@ static NTSTATUS pdb_wbc_sam_getsampwnam(struct pdb_methods *methods, struct samu
 	return _pdb_wbc_sam_getsampw(methods, user, winbind_getpwnam(sname));
 }
 
-static NTSTATUS pdb_wbc_sam_getsampwsid(struct pdb_methods *methods, struct samu *user, const DOM_SID *sid)
+static NTSTATUS pdb_wbc_sam_getsampwsid(struct pdb_methods *methods, struct samu *user, const struct dom_sid *sid)
 {
 	return _pdb_wbc_sam_getsampw(methods, user, winbind_getpwsid(sid));
 }
 
 static bool pdb_wbc_sam_uid_to_sid(struct pdb_methods *methods, uid_t uid,
-				   DOM_SID *sid)
+				   struct dom_sid *sid)
 {
 	return winbind_uid_to_sid(sid, uid);
 }
 
 static bool pdb_wbc_sam_gid_to_sid(struct pdb_methods *methods, gid_t gid,
-				   DOM_SID *sid)
+				   struct dom_sid *sid)
 {
 	return winbind_gid_to_sid(sid, gid);
 }
 
 static NTSTATUS pdb_wbc_sam_enum_group_members(struct pdb_methods *methods,
 					       TALLOC_CTX *mem_ctx,
-					       const DOM_SID *group,
+					       const struct dom_sid *group,
 					       uint32 **pp_member_rids,
 					       size_t *p_num_members)
 {
@@ -93,9 +95,9 @@ static NTSTATUS pdb_wbc_sam_enum_group_members(struct pdb_methods *methods,
 static NTSTATUS pdb_wbc_sam_enum_group_memberships(struct pdb_methods *methods,
 						   TALLOC_CTX *mem_ctx,
 						   struct samu *user,
-						   DOM_SID **pp_sids,
+						   struct dom_sid **pp_sids,
 						   gid_t **pp_gids,
-						   size_t *p_num_groups)
+						   uint32_t *p_num_groups)
 {
 	size_t i;
 	const char *username = pdb_get_username(user);
@@ -110,7 +112,7 @@ static NTSTATUS pdb_wbc_sam_enum_group_memberships(struct pdb_methods *methods,
 		smb_panic("primary group missing");
 	}
 
-	*pp_sids = TALLOC_ARRAY(mem_ctx, DOM_SID, *p_num_groups);
+	*pp_sids = TALLOC_ARRAY(mem_ctx, struct dom_sid, *p_num_groups);
 
 	if (*pp_sids == NULL) {
 		TALLOC_FREE(*pp_gids);
@@ -125,7 +127,7 @@ static NTSTATUS pdb_wbc_sam_enum_group_memberships(struct pdb_methods *methods,
 }
 
 static NTSTATUS pdb_wbc_sam_lookup_rids(struct pdb_methods *methods,
-					const DOM_SID *domain_sid,
+					const struct dom_sid *domain_sid,
 					int num_rids,
 					uint32 *rids,
 					const char **names,
@@ -185,7 +187,7 @@ static bool pdb_wbc_sam_search_groups(struct pdb_methods *methods,
 
 static bool pdb_wbc_sam_search_aliases(struct pdb_methods *methods,
 				       struct pdb_search *search,
-				       const DOM_SID *sid)
+				       const struct dom_sid *sid)
 {
 
 	return false;
@@ -194,7 +196,7 @@ static bool pdb_wbc_sam_search_aliases(struct pdb_methods *methods,
 static bool pdb_wbc_sam_get_trusteddom_pw(struct pdb_methods *methods,
 					  const char *domain,
 					  char **pwd,
-					  DOM_SID *sid,
+					  struct dom_sid *sid,
 					  time_t *pass_last_set_time)
 {
 	return false;
@@ -204,7 +206,7 @@ static bool pdb_wbc_sam_get_trusteddom_pw(struct pdb_methods *methods,
 static bool pdb_wbc_sam_set_trusteddom_pw(struct pdb_methods *methods,
 					  const char *domain,
 					  const char *pwd,
-					  const DOM_SID *sid)
+					  const struct dom_sid *sid)
 {
 	return false;
 }
@@ -223,7 +225,7 @@ static NTSTATUS pdb_wbc_sam_enum_trusteddoms(struct pdb_methods *methods,
 	return NT_STATUS_NOT_IMPLEMENTED;
 }
 
-static bool _make_group_map(struct pdb_methods *methods, const char *domain, const char *name, enum lsa_SidType name_type, gid_t gid, DOM_SID *sid, GROUP_MAP *map)
+static bool _make_group_map(struct pdb_methods *methods, const char *domain, const char *name, enum lsa_SidType name_type, gid_t gid, struct dom_sid *sid, GROUP_MAP *map)
 {
 	snprintf(map->nt_name, sizeof(map->nt_name), "%s%c%s",
 	        domain, *lp_winbind_separator(), name);
@@ -234,7 +236,7 @@ static bool _make_group_map(struct pdb_methods *methods, const char *domain, con
 }
 
 static NTSTATUS pdb_wbc_sam_getgrsid(struct pdb_methods *methods, GROUP_MAP *map,
-				 DOM_SID sid)
+				 struct dom_sid sid)
 {
 	NTSTATUS result = NT_STATUS_OK;
 	char *name = NULL;
@@ -278,7 +280,7 @@ static NTSTATUS pdb_wbc_sam_getgrgid(struct pdb_methods *methods, GROUP_MAP *map
 	NTSTATUS result = NT_STATUS_OK;
 	char *name = NULL;
 	char *domain = NULL;
-	DOM_SID sid;
+	struct dom_sid sid;
 	enum lsa_SidType name_type;
 
 	if (!winbind_gid_to_sid(&sid, gid)) {
@@ -317,7 +319,7 @@ static NTSTATUS pdb_wbc_sam_getgrnam(struct pdb_methods *methods, GROUP_MAP *map
 {
 	NTSTATUS result = NT_STATUS_OK;
 	const char *domain = "";
-	DOM_SID sid;
+	struct dom_sid sid;
 	gid_t gid;
 	enum lsa_SidType name_type;
 
@@ -350,7 +352,7 @@ done:
 }
 
 static NTSTATUS pdb_wbc_sam_enum_group_mapping(struct pdb_methods *methods,
-					   const DOM_SID *sid, enum lsa_SidType sid_name_use,
+					   const struct dom_sid *sid, enum lsa_SidType sid_name_use,
 					   GROUP_MAP **pp_rmap, size_t *p_num_entries,
 					   bool unix_only)
 {
@@ -358,16 +360,16 @@ static NTSTATUS pdb_wbc_sam_enum_group_mapping(struct pdb_methods *methods,
 }
 
 static NTSTATUS pdb_wbc_sam_get_aliasinfo(struct pdb_methods *methods,
-				   const DOM_SID *sid,
+				   const struct dom_sid *sid,
 				   struct acct_info *info)
 {
 	return NT_STATUS_NOT_IMPLEMENTED;
 }
 
 static NTSTATUS pdb_wbc_sam_enum_aliasmem(struct pdb_methods *methods,
-					  const DOM_SID *alias,
+					  const struct dom_sid *alias,
 					  TALLOC_CTX *mem_ctx,
-					  DOM_SID **pp_members,
+					  struct dom_sid **pp_members,
 					  size_t *p_num_members)
 {
 	return NT_STATUS_NOT_IMPLEMENTED;
@@ -375,8 +377,8 @@ static NTSTATUS pdb_wbc_sam_enum_aliasmem(struct pdb_methods *methods,
 
 static NTSTATUS pdb_wbc_sam_alias_memberships(struct pdb_methods *methods,
 				       TALLOC_CTX *mem_ctx,
-				       const DOM_SID *domain_sid,
-				       const DOM_SID *members,
+				       const struct dom_sid *domain_sid,
+				       const struct dom_sid *members,
 				       size_t num_members,
 				       uint32 **pp_alias_rids,
 				       size_t *p_num_alias_rids)
