@@ -1656,6 +1656,28 @@ static void add_child_pid(pid_t pid)
         num_children += 1;
 }
 
+static bool printer_housekeeping_fn(const struct timeval *now,
+				    void *private_data)
+{
+	static time_t last_pcap_reload_time = 0;
+	time_t printcap_cache_time = (time_t)lp_printcap_cache_time();
+	time_t t = time_mono(NULL);
+
+	DEBUG(5, ("printer housekeeping\n"));
+
+	/* if periodic printcap rescan is enabled, see if it's time to reload */
+	if ((printcap_cache_time != 0)
+	 && (t >= (last_pcap_reload_time + printcap_cache_time))) {
+		DEBUG( 3,( "Printcap cache time expired.\n"));
+		pcap_cache_reload(server_event_context(),
+				  smbd_messaging_context(),
+				  &reload_pcap_change_notify);
+		last_pcap_reload_time = t;
+	}
+
+	return true;
+}
+
 static pid_t background_lpq_updater_pid = -1;
 
 /****************************************************************************
@@ -1729,6 +1751,15 @@ void start_background_queue(struct tevent_context *ev,
 			smb_panic("tevent_add_fd() failed for pause_pipe");
 		}
 
+		if (!(event_add_idle(ev, NULL,
+				     timeval_set(SMBD_HOUSEKEEPING_INTERVAL, 0),
+				     "printer_housekeeping",
+				     printer_housekeeping_fn,
+				     NULL))) {
+			DEBUG(0, ("Could not add printing housekeeping event\n"));
+			exit(1);
+		}
+
 		DEBUG(5,("start_background_queue: background LPQ thread waiting for messages\n"));
 		ret = tevent_loop_wait(ev);
 		/* should not be reached */
@@ -2843,8 +2874,7 @@ static WERROR print_job_spool_file(int snum, uint32_t jobid,
 	}
 
 	slprintf(pjob->filename, sizeof(pjob->filename)-1,
-		 "%s/%s%.8u.XXXXXX", lp_pathname(snum),
-		 PRINT_SPOOL_PREFIX, (unsigned int)jobid);
+		 "%s/%sXXXXXX", lp_pathname(snum), PRINT_SPOOL_PREFIX);
 	pjob->fd = mkstemp(pjob->filename);
 
 	if (pjob->fd == -1) {
