@@ -29,13 +29,13 @@
 static struct tevent_req *smbd_smb2_getinfo_send(TALLOC_CTX *mem_ctx,
 						 struct tevent_context *ev,
 						 struct smbd_smb2_request *smb2req,
+						 struct files_struct *in_fsp,
 						 uint8_t in_info_type,
 						 uint8_t in_file_info_class,
 						 uint32_t in_output_buffer_length,
 						 DATA_BLOB in_input_buffer,
 						 uint32_t in_additional_information,
-						 uint32_t in_flags,
-						 uint64_t in_file_id_volatile);
+						 uint32_t in_flags);
 static NTSTATUS smbd_smb2_getinfo_recv(struct tevent_req *req,
 				       TALLOC_CTX *mem_ctx,
 				       DATA_BLOB *out_output_buffer,
@@ -57,6 +57,7 @@ NTSTATUS smbd_smb2_request_process_getinfo(struct smbd_smb2_request *req)
 	uint32_t in_flags;
 	uint64_t in_file_id_persistent;
 	uint64_t in_file_id_volatile;
+	struct files_struct *in_fsp;
 	struct tevent_req *subreq;
 
 	status = smbd_smb2_request_verify_sizes(req, 0x29);
@@ -97,22 +98,19 @@ NTSTATUS smbd_smb2_request_process_getinfo(struct smbd_smb2_request *req)
 		return smbd_smb2_request_error(req, NT_STATUS_INVALID_PARAMETER);
 	}
 
-	if (req->compat_chain_fsp) {
-		/* skip check */
-	} else if (in_file_id_persistent != in_file_id_volatile) {
+	in_fsp = file_fsp_smb2(req, in_file_id_persistent, in_file_id_volatile);
+	if (in_fsp == NULL) {
 		return smbd_smb2_request_error(req, NT_STATUS_FILE_CLOSED);
 	}
 
-	subreq = smbd_smb2_getinfo_send(req,
-					req->sconn->smb2.event_ctx,
-					req,
+	subreq = smbd_smb2_getinfo_send(req, req->sconn->smb2.event_ctx,
+					req, in_fsp,
 					in_info_type,
 					in_file_info_class,
 					in_output_buffer_length,
 					in_input_buffer,
 					in_additional_information,
-					in_flags,
-					in_file_id_volatile);
+					in_flags);
 	if (subreq == NULL) {
 		return smbd_smb2_request_error(req, NT_STATUS_NO_MEMORY);
 	}
@@ -234,19 +232,18 @@ static void smb2_ipc_getinfo(struct tevent_req *req,
 static struct tevent_req *smbd_smb2_getinfo_send(TALLOC_CTX *mem_ctx,
 						 struct tevent_context *ev,
 						 struct smbd_smb2_request *smb2req,
+						 struct files_struct *fsp,
 						 uint8_t in_info_type,
 						 uint8_t in_file_info_class,
 						 uint32_t in_output_buffer_length,
 						 DATA_BLOB in_input_buffer,
 						 uint32_t in_additional_information,
-						 uint32_t in_flags,
-						 uint64_t in_file_id_volatile)
+						 uint32_t in_flags)
 {
 	struct tevent_req *req;
 	struct smbd_smb2_getinfo_state *state;
 	struct smb_request *smbreq;
 	connection_struct *conn = smb2req->tcon->compat_conn;
-	files_struct *fsp;
 	NTSTATUS status;
 
 	req = tevent_req_create(mem_ctx, &state,
@@ -258,28 +255,14 @@ static struct tevent_req *smbd_smb2_getinfo_send(TALLOC_CTX *mem_ctx,
 	state->status = NT_STATUS_OK;
 	state->out_output_buffer = data_blob_null;
 
-	DEBUG(10,("smbd_smb2_getinfo_send: file_id[0x%016llX]\n",
-		  (unsigned long long)in_file_id_volatile));
+	DEBUG(10,("smbd_smb2_getinfo_send: %s - fnum[%d]\n",
+		  fsp_str_dbg(fsp), fsp->fnum));
 
 	smbreq = smbd_smb2_fake_smb_request(smb2req);
 	if (tevent_req_nomem(smbreq, req)) {
 		return tevent_req_post(req, ev);
 	}
 
-	fsp = file_fsp(smbreq, (uint16_t)in_file_id_volatile);
-	if (fsp == NULL) {
-		tevent_req_nterror(req, NT_STATUS_FILE_CLOSED);
-		return tevent_req_post(req, ev);
-	}
-	if (conn != fsp->conn) {
-		tevent_req_nterror(req, NT_STATUS_FILE_CLOSED);
-		return tevent_req_post(req, ev);
-	}
-	if (smb2req->session->vuid != fsp->vuid) {
-		tevent_req_nterror(req, NT_STATUS_FILE_CLOSED);
-		return tevent_req_post(req, ev);
-	}
-
 	if (IS_IPC(conn)) {
 		smb2_ipc_getinfo(req, state, ev,
 			in_info_type, in_file_info_class);
