@@ -8,22 +8,30 @@
    Copyright (C) Martin Pool 2001-2002
    Copyright (C) Simo Sorce 2002
    Copyright (C) James Peach 2006
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "includes.h"
+#include "system/filesys.h"
+#include "popt_common.h"
+#include "librpc/gen_ndr/spoolss.h"
+#include "nt_printing.h"
+#include "printing/notify.h"
+#include "libsmb/nmblib.h"
+#include "messages.h"
+#include "util_tdb.h"
 
 #if HAVE_LIBUNWIND_H
 #include <libunwind.h>
@@ -61,7 +69,7 @@ static bool send_message(struct messaging_context *msg_ctx,
 	ret = message_send_all(msg_ctx, msg_type, buf, len, &n_sent);
 	DEBUG(10,("smbcontrol/send_message: broadcast message to "
 		  "%d processes\n", n_sent));
-	
+
 	return ret;
 }
 
@@ -109,8 +117,12 @@ static void print_pid_string_cb(struct messaging_context *msg,
 				struct server_id pid,
 				DATA_BLOB *data)
 {
-	printf("PID %u: %.*s", (unsigned int)procid_to_pid(&pid),
-	       (int)data->length, (const char *)data->data);
+	char *pidstr;
+
+	pidstr = procid_str(talloc_tos(), &pid);
+	printf("PID %s: %.*s", pidstr, (int)data->length,
+	       (const char *)data->data);
+	TALLOC_FREE(pidstr);
 	num_replies++;
 }
 
@@ -122,7 +134,7 @@ static void print_string_cb(struct messaging_context *msg,
 			    struct server_id pid,
 			    DATA_BLOB *data)
 {
-	printf("%.*s", (int)data->length, (const char *)data->data);
+	printf("%*s", (int)data->length, (const char *)data->data);
 	num_replies++;
 }
 
@@ -158,6 +170,54 @@ static bool do_debug(struct messaging_context *msg_ctx,
 			    strlen(argv[1]) + 1);
 }
 
+
+static bool do_idmap(struct messaging_context *msg_ctx,
+		     const struct server_id pid,
+		     const int argc, const char **argv)
+{
+	static const char* usage = "Usage: "
+		"smbcontrol <dest> idmap <cmd> [arg]\n"
+		"\tcmd:\tflush [gid|uid]\n"
+		"\t\tdelete \"UID <uid>\"|\"GID <gid>\"|<sid>\n"
+		"\t\tkill \"UID <uid>\"|\"GID <gid>\"|<sid>\n";
+	const char* arg = NULL;
+	int arglen = 0;
+	int msg_type;
+
+	switch (argc) {
+	case 2:
+		break;
+	case 3:
+		arg = argv[2];
+		arglen = strlen(arg) + 1;
+		break;
+	default:
+		fprintf(stderr, "%s", usage);
+		return false;
+	}
+
+	if (strcmp(argv[1], "flush") == 0) {
+		msg_type = MSG_IDMAP_FLUSH;
+	}
+	else if (strcmp(argv[1], "delete") == 0) {
+		msg_type = MSG_IDMAP_DELETE;
+	}
+	else if (strcmp(argv[1], "kill") == 0) {
+		msg_type = MSG_IDMAP_KILL;
+	}
+	else if (strcmp(argv[1], "help") == 0) {
+		fprintf(stdout, "%s", usage);
+		return true;
+	}
+	else {
+		fprintf(stderr, "%s", usage);
+		return false;
+	}
+
+	return send_message(msg_ctx, pid, msg_type, arg, arglen);
+}
+
+
 #if defined(HAVE_LIBUNWIND_PTRACE) && defined(HAVE_LINUX_PTRACE)
 
 /* Return the name of a process given it's PID. This will only work on Linux,
@@ -259,8 +319,7 @@ cleanup:
 	ptrace(PTRACE_DETACH, pid, NULL, NULL);
 }
 
-static int stack_trace_connection(struct db_record *rec,
-				  const struct connections_key *key,
+static int stack_trace_connection(const struct connections_key *key,
 				  const struct connections_data *crec,
 				  void *priv)
 {
@@ -291,7 +350,7 @@ static bool do_daemon_stack_trace(struct messaging_context *msg_ctx,
 		 */
 		print_stack_trace(dest, &count);
 	} else {
-		connections_forall(stack_trace_connection, &count);
+		connections_forall_read(stack_trace_connection, &count);
 	}
 
 	return True;
@@ -476,7 +535,7 @@ static void profilelevel_cb(struct messaging_context *msg_ctx,
 		s = "BOGUS";
 		break;
 	}
-	
+
 	printf("Profiling %s on pid %u\n",s,(unsigned int)procid_to_pid(&pid));
 }
 
@@ -572,7 +631,7 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 		fprintf(stderr, "\tjobdelete <printername> <unix jobid>\n");
 		fprintf(stderr, "\tprinter <printername> <comment|port|"
 			"driver> <value>\n");
-		
+
 		return False;
 	}
 
@@ -585,8 +644,10 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 				" queuepause <printername>\n");
 			return False;
 		}
-		
-		notify_printer_status_byname(argv[2], PRINTER_STATUS_PAUSED);
+
+		notify_printer_status_byname(messaging_event_context(msg_ctx),
+					     msg_ctx, argv[2],
+					     PRINTER_STATUS_PAUSED);
 
 		goto send;
 
@@ -597,8 +658,10 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 				" queuereume <printername>\n");
 			return False;
 		}
-		
-		notify_printer_status_byname(argv[2], PRINTER_STATUS_OK);
+
+		notify_printer_status_byname(messaging_event_context(msg_ctx),
+					     msg_ctx, argv[2],
+					     PRINTER_STATUS_OK);
 
 		goto send;
 
@@ -614,7 +677,8 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 		jobid = atoi(argv[3]);
 
 		notify_job_status_byname(
-			argv[2], jobid, JOB_STATUS_PAUSED, 
+			messaging_event_context(msg_ctx), msg_ctx,
+			argv[2], jobid, JOB_STATUS_PAUSED,
 			SPOOLSS_NOTIFY_MSG_UNIX_JOBID);
 
 		goto send;
@@ -631,6 +695,7 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 		jobid = atoi(argv[3]);
 
 		notify_job_status_byname(
+			messaging_event_context(msg_ctx), msg_ctx,
 			argv[2], jobid, JOB_STATUS_QUEUED, 
 			SPOOLSS_NOTIFY_MSG_UNIX_JOBID);
 
@@ -648,10 +713,12 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 		jobid = atoi(argv[3]);
 
 		notify_job_status_byname(
+			messaging_event_context(msg_ctx), msg_ctx,
 			argv[2], jobid, JOB_STATUS_DELETING,
 			SPOOLSS_NOTIFY_MSG_UNIX_JOBID);
-		
+
 		notify_job_status_byname(
+			messaging_event_context(msg_ctx), msg_ctx,
 			argv[2], jobid, JOB_STATUS_DELETING|
 			JOB_STATUS_DELETED,
 			SPOOLSS_NOTIFY_MSG_UNIX_JOBID);
@@ -660,7 +727,7 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 
 	} else if (strcmp(cmd, "printer") == 0) {
 		uint32 attribute;
-		
+
 		if (argc != 5) {
 			fprintf(stderr, "Usage: smbcontrol <dest> printnotify "
 				"printer <printername> <comment|port|driver> "
@@ -680,7 +747,8 @@ static bool do_printnotify(struct messaging_context *msg_ctx,
 			return False;
 		}
 
-		notify_printer_byname(argv[2], attribute,
+		notify_printer_byname(messaging_event_context(msg_ctx),
+				      msg_ctx, argv[2], attribute,
 				      CONST_DISCARD(char *, argv[4]));
 
 		goto send;
@@ -710,6 +778,22 @@ static bool do_closeshare(struct messaging_context *msg_ctx,
 			    strlen(argv[1]) + 1);
 }
 
+/* Tell winbindd an IP got dropped */
+
+static bool do_ip_dropped(struct messaging_context *msg_ctx,
+			  const struct server_id pid,
+			  const int argc, const char **argv)
+{
+	if (argc != 2) {
+		fprintf(stderr, "Usage: smbcontrol <dest> ip-dropped "
+			"<ip-address>\n");
+		return False;
+	}
+
+	return send_message(msg_ctx, pid, MSG_WINBIND_IP_DROPPED, argv[1],
+			    strlen(argv[1]) + 1);
+}
+
 /* force a blocking lock retry */
 
 static bool do_lockretry(struct messaging_context *msg_ctx,
@@ -902,7 +986,8 @@ static bool do_winbind_offline(struct messaging_context *msg_ctx,
 
 	tdb = tdb_open_log(cache_path("winbindd_cache.tdb"),
 				WINBINDD_CACHE_TDB_DEFAULT_HASH_SIZE,
-				TDB_DEFAULT /* TDB_CLEAR_IF_FIRST */, O_RDWR|O_CREAT, 0600);
+				TDB_DEFAULT|TDB_INCOMPATIBLE_HASH /* TDB_CLEAR_IF_FIRST */,
+				O_RDWR|O_CREAT, 0600);
 
 	if (!tdb) {
 		fprintf(stderr, "Cannot open the tdb %s for writing.\n",
@@ -934,7 +1019,7 @@ static bool do_winbind_offline(struct messaging_context *msg_ctx,
 
 		/* Check that the entry "WINBINDD_OFFLINE" still exists. */
 		d = tdb_fetch_bystring( tdb, "WINBINDD_OFFLINE" );
-	
+
 		if (!d.dptr || d.dsize != 4) {
 			SAFE_FREE(d.dptr);
 			DEBUG(10,("do_winbind_offline: offline state not set - retrying.\n"));
@@ -954,7 +1039,7 @@ static bool do_winbind_onlinestatus(struct messaging_context *msg_ctx,
 {
 	struct server_id myid;
 
-	myid = pid_to_procid(sys_getpid());
+	myid = messaging_server_id(msg_ctx);
 
 	if (argc != 1) {
 		fprintf(stderr, "Usage: smbcontrol winbindd onlinestatus\n");
@@ -986,7 +1071,7 @@ static bool do_dump_event_list(struct messaging_context *msg_ctx,
 {
 	struct server_id myid;
 
-	myid = pid_to_procid(sys_getpid());
+	myid = messaging_server_id(msg_ctx);
 
 	if (argc != 1) {
 		fprintf(stderr, "Usage: smbcontrol <dest> dump-event-list\n");
@@ -1006,10 +1091,10 @@ static bool do_winbind_dump_domain_list(struct messaging_context *msg_ctx,
 	uint8_t *buf = NULL;
 	int buf_len = 0;
 
-	myid = pid_to_procid(sys_getpid());
+	myid = messaging_server_id(msg_ctx);
 
 	if (argc < 1 || argc > 2) {
-		fprintf(stderr, "Usage: smbcontrol <dest> dump_domain_list "
+		fprintf(stderr, "Usage: smbcontrol <dest> dump-domain-list "
 			"<domain>\n");
 		return false;
 	}
@@ -1069,7 +1154,9 @@ static bool do_winbind_validate_cache(struct messaging_context *msg_ctx,
 				      const struct server_id pid,
 				      const int argc, const char **argv)
 {
-	struct server_id myid = pid_to_procid(sys_getpid());
+	struct server_id myid;
+
+	myid = messaging_server_id(msg_ctx);
 
 	if (argc != 1) {
 		fprintf(stderr, "Usage: smbcontrol winbindd validate-cache\n");
@@ -1165,6 +1252,7 @@ static const struct {
 	const char *help;	/* Short help text */
 } msg_types[] = {
 	{ "debug", do_debug, "Set debuglevel"  },
+	{ "idmap", do_idmap, "Manipulate idmap cache" },
 	{ "force-election", do_election,
 	  "Force a browse election" },
 	{ "ping", do_ping, "Elicit a response" },
@@ -1177,6 +1265,7 @@ static const struct {
 	{ "debuglevel", do_debuglevel, "Display current debuglevels" },
 	{ "printnotify", do_printnotify, "Send a print notify message" },
 	{ "close-share", do_closeshare, "Forcibly disconnect a share" },
+	{ "ip-dropped", do_ip_dropped, "Tell winbind that an IP got dropped" },
 	{ "lockretry", do_lockretry, "Force a blocking lock retry" },
 	{ "brl-revalidate", do_brl_revalidate, "Revalidate all brl entries" },
         { "samsync", do_samsync, "Initiate SAM synchronisation" },
@@ -1225,7 +1314,8 @@ static void usage(poptContext pc)
 
 /* Return the pid number for a string destination */
 
-static struct server_id parse_dest(const char *dest)
+static struct server_id parse_dest(struct messaging_context *msg,
+				   const char *dest)
 {
 	struct server_id result = {-1};
 	pid_t pid;
@@ -1239,7 +1329,7 @@ static struct server_id parse_dest(const char *dest)
 	/* Try self - useful for testing */
 
 	if (strequal(dest, "self")) {
-		return pid_to_procid(sys_getpid());
+		return messaging_server_id(msg);
 	}
 
 	/* Fix winbind typo. */
@@ -1277,7 +1367,7 @@ static bool do_command(struct messaging_context *msg_ctx,
 
 	/* Check destination */
 
-	pid = parse_dest(dest);
+	pid = parse_dest(msg_ctx, dest);
 	if (!procid_valid(&pid)) {
 		return False;
 	}
@@ -1342,8 +1432,8 @@ int main(int argc, const char **argv)
 
 	load_case_tables();
 
-	setup_logging(argv[0],True);
-	
+	setup_logging(argv[0], DEBUG_STDOUT);
+
 	/* Parse command line arguments using popt */
 
 	pc = poptGetContext(
@@ -1386,14 +1476,14 @@ int main(int argc, const char **argv)
 	/* Need to invert sense of return code -- samba
          * routines mostly return True==1 for success, but
          * shell needs 0. */ 
-	
+
 	if (!(evt_ctx = tevent_context_init(NULL)) ||
-	    !(msg_ctx = messaging_init(NULL, server_id_self(), evt_ctx))) {
+	    !(msg_ctx = messaging_init(NULL, procid_self(), evt_ctx))) {
 		fprintf(stderr, "could not init messaging context\n");
 		TALLOC_FREE(frame);
 		exit(1);
 	}
-	
+
 	ret = !do_command(msg_ctx, argc, argv);
 	TALLOC_FREE(frame);
 	return ret;
