@@ -4,17 +4,17 @@
    Version 3.0.0
    Copyright (C) Andrew Tridgell 1997-2002
    Copyright (C) John H Terpstra 2002
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
@@ -28,7 +28,13 @@
  **/
 
 #include "includes.h"
+#include "system/filesys.h"
+#include "popt_common.h"
 #include "web/swat_proto.h"
+#include "printing/pcap.h"
+#include "printing/load.h"
+#include "passdb.h"
+#include "intl/lang_tdb.h"
 #include "../lib/crypto/md5.h"
 
 static int demo_mode = False;
@@ -121,7 +127,7 @@ static char *stripspaceupper(const char *str)
 	char *p = newstring;
 
 	while (*str) {
-		if (*str != ' ') *p++ = toupper_ascii(*str);
+		if (*str != ' ') *p++ = toupper_m(*str);
 		++str;
 	}
 	*p = '\0';
@@ -145,9 +151,10 @@ static char *make_parm_name(const char *label)
 void get_xsrf_token(const char *username, const char *pass,
 		    const char *formname, time_t xsrf_time, char token_str[33])
 {
-	struct MD5Context md5_ctx;
+	MD5_CTX md5_ctx;
 	uint8_t token[16];
 	int i;
+	char *nonce = cgi_nonce();
 
 	token_str[0] = '\0';
 	ZERO_STRUCT(md5_ctx);
@@ -161,6 +168,7 @@ void get_xsrf_token(const char *username, const char *pass,
 	if (pass != NULL) {
 		MD5Update(&md5_ctx, (uint8_t *)pass, strlen(pass));
 	}
+	MD5Update(&md5_ctx, (uint8_t *)nonce, strlen(nonce));
 
 	MD5Final(token, &md5_ctx);
 
@@ -168,7 +176,7 @@ void get_xsrf_token(const char *username, const char *pass,
 		char tmp[3];
 
 		snprintf(tmp, sizeof(tmp), "%02x", token[i]);
-		strncat(token_str, tmp, sizeof(tmp));
+		strlcat(token_str, tmp, sizeof(tmp));
 	}
 }
 
@@ -192,16 +200,29 @@ bool verify_xsrf_token(const char *formname)
 	const char *pass = cgi_user_pass();
 	const char *token = cgi_variable_nonull(XSRF_TOKEN);
 	const char *time_str = cgi_variable_nonull(XSRF_TIME);
+	char *p = NULL;
+	long long xsrf_time_ll = 0;
 	time_t xsrf_time = 0;
 	time_t now = time(NULL);
 
-	if (sizeof(time_t) == sizeof(int)) {
-		xsrf_time = atoi(time_str);
-	} else if (sizeof(time_t) == sizeof(long)) {
-		xsrf_time = atol(time_str);
-	} else if (sizeof(time_t) == sizeof(long long)) {
-		xsrf_time = atoll(time_str);
+	errno = 0;
+	xsrf_time_ll = strtoll(time_str, &p, 10);
+	if (errno != 0) {
+		return false;
 	}
+	if (p == NULL) {
+		return false;
+	}
+	if (PTR_DIFF(p, time_str) > strlen(time_str)) {
+		return false;
+	}
+	if (xsrf_time_ll > _TYPE_MAXIMUM(time_t)) {
+		return false;
+	}
+	if (xsrf_time_ll < _TYPE_MINIMUM(time_t)) {
+		return false;
+	}
+	xsrf_time = xsrf_time_ll;
 
 	if (abs(now - xsrf_time) > XSRF_TIMEOUT) {
 		return false;
@@ -247,7 +268,8 @@ static void print_header(void)
 	if (!cgi_waspost()) {
 		printf("Expires: 0\r\n");
 	}
-	printf("Content-type: text/html\r\n\r\n");
+	printf("Content-type: text/html\r\n");
+	printf("X-Frame-Options: DENY\r\n\r\n");
 
 	if (!include_html("include/header.html")) {
 		printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n");
@@ -491,9 +513,9 @@ static void show_parameters(int snum, int allparameters, unsigned int parm_filte
 		}
 
 		if ((parm_filter & FLAG_WIZARD) && !(parm->flags & FLAG_WIZARD)) continue;
-		
+
 		if ((parm_filter & FLAG_ADVANCED) && !(parm->flags & FLAG_ADVANCED)) continue;
-		
+
 		if (heading && heading != last_heading) {
 			printf("<tr><td></td></tr><tr><td><b><u>%s</u></b></td></tr>\n", _(heading));
 			last_heading = heading;
@@ -520,7 +542,7 @@ static void write_config(FILE *f, bool show_defaults)
 	fprintf(f, "# Samba config file created using SWAT\n");
 	fprintf(f, "# from %s (%s)\n", cgi_remote_host(), cgi_remote_addr());
 	fprintf(f, "# Date: %s\n\n", current_timestring(ctx, False));
-	
+
 	lp_dump(f, show_defaults, iNumNonAutoPrintServices);
 
 	TALLOC_FREE(ctx);
@@ -564,7 +586,10 @@ static int save_reload(int snum)
                 return 0;
         }
 	iNumNonAutoPrintServices = lp_numservices();
-	pcap_cache_reload(&load_printers);
+	if (pcap_cache_loaded()) {
+		load_printers(server_event_context(),
+			      server_messaging_context());
+	}
 
 	return 1;
 }
@@ -630,7 +655,7 @@ static void image_link(const char *name, const char *hlink, const char *src)
 static void show_main_buttons(void)
 {
 	char *p;
-	
+
 	if ((p = cgi_user_name()) && strcmp(p, "root")) {
 		printf(_("Logged in as <b>%s</b>"), p);
 		printf("<p>\n");
@@ -744,7 +769,7 @@ output_page:
 
 	printf("<input type=reset name=\"Reset Values\" value=\"Reset\">\n");
 	printf("<p>\n");
-	
+
 	printf("<table>\n");
 	show_parameters(GLOBAL_SECTION_SNUM, 1, parm_filter, 0);
 	printf("</table>\n");
@@ -796,7 +821,7 @@ static void wizard_page(void)
 
 		/* Plain text passwords are too badly broken - use encrypted passwords only */
 		lp_do_parameter( GLOBAL_SECTION_SNUM, "encrypt passwords", "Yes");
-		
+
 		switch ( SerType ){
 			case 0:
 				/* Stand-alone Server */
@@ -908,7 +933,7 @@ output_page:
 		const char **wins_servers = lp_wins_server_list();
 		for(i = 0; wins_servers[i]; i++) printf("%s ", wins_servers[i]);
 	}
-	
+
 	printf("\"></td></tr>\n");
 	if (winstype == 3) {
 		printf("<tr><td></td><td colspan=3><font color=\"#ff0000\">%s</font></td></tr>\n", _("Error: WINS Server Mode and WINS Support both set in smb.conf"));
@@ -918,14 +943,14 @@ output_page:
 	printf("<td><input type=radio name=\"HomeExpo\" value=\"1\" %s> Yes</td>", (have_home == -1) ? "" : "checked ");
 	printf("<td><input type=radio name=\"HomeExpo\" value=\"0\" %s> No</td>", (have_home == -1 ) ? "checked" : "");
 	printf("<td></td></tr>\n");
-	
+
 	/* Enable this when we are ready ....
 	 * printf("<tr><td><b>%s:&nbsp;</b></td>\n", _("Is Print Server"));
 	 * printf("<td><input type=radio name=\"PtrSvr\" value=\"1\" %s> Yes</td>");
 	 * printf("<td><input type=radio name=\"PtrSvr\" value=\"0\" %s> No</td>");
 	 * printf("<td></td></tr>\n");
 	 */
-	
+
 	printf("</table></center>");
 	printf("<hr>");
 
@@ -1124,7 +1149,7 @@ static bool change_password(const char *remote_machine, const char *user_name,
 		printf("%s\n<p>", _("password change in demo mode rejected"));
 		return False;
 	}
-	
+
 	if (remote_machine != NULL) {
 		ret = remote_password_change(remote_machine, user_name,
 					     old_passwd, new_passwd, &err_str);
@@ -1138,7 +1163,7 @@ static bool change_password(const char *remote_machine, const char *user_name,
 		printf("%s\n<p>", _("Can't setup password database vectors."));
 		return False;
 	}
-	
+
 	ret = local_password_change(user_name, local_flags, new_passwd,
 					&err_str, &msg_str);
 
@@ -1221,7 +1246,6 @@ static void chg_passwd(void)
 	local_flags |= (cgi_variable(DELETE_USER_FLAG) ? LOCAL_DELETE_USER : 0);
 	local_flags |= (cgi_variable(ENABLE_USER_FLAG) ? LOCAL_ENABLE_USER : 0);
 	local_flags |= (cgi_variable(DISABLE_USER_FLAG) ? LOCAL_DISABLE_USER : 0);
-	
 
 	rslt = change_password(host,
 			       cgi_variable_nonull(SWAT_USER),
@@ -1236,7 +1260,7 @@ static void chg_passwd(void)
 			printf("%s\n", _(" The passwd has NOT been changed."));
 		}
 	}
-	
+
 	return;
 }
 
@@ -1526,13 +1550,15 @@ const char *lang_msg_rotate(TALLOC_CTX *ctx, const char *msgid)
 	/* we don't want any SIGPIPE messages */
 	BlockSignals(True,SIGPIPE);
 
-	dbf = x_fopen("/dev/null", O_WRONLY, 0);
-	if (!dbf) dbf = x_stderr;
+	debug_set_logfile("/dev/null");
 
 	/* we don't want stderr screwing us up */
 	close(2);
 	open("/dev/null", O_WRONLY);
+	setup_logging("swat", DEBUG_FILE);
 
+	load_case_tables();
+	
 	pc = poptGetContext("swat", argc, (const char **) argv, long_options, 0);
 
 	/* Parse command line options */
@@ -1541,13 +1567,15 @@ const char *lang_msg_rotate(TALLOC_CTX *ctx, const char *msgid)
 
 	poptFreeContext(pc);
 
-	load_case_tables();
-
-	setup_logging(argv[0],False);
+	/* This should set a more apporiate log file */
 	load_config(True);
+	reopen_logs();
 	load_interfaces();
 	iNumNonAutoPrintServices = lp_numservices();
-	pcap_cache_reload(&load_printers);
+	if (pcap_cache_loaded()) {
+		load_printers(server_event_context(),
+			      server_messaging_context());
+	}
 
 	cgi_setup(get_dyn_SWATDIR(), !demo_mode);
 
